from pathlib import Path

import customtkinter as ctk

from pygcg.utils.misc import fpe


class SettingsSelection(ctk.CTkFrame):
    def __init__(
        self, master, row, label, value, setting_is_dir=False, *args, **kwargs
    ):
        super().__init__(master, *args, **kwargs)

        self.value_key = value
        self.setting_is_dir = setting_is_dir

        self.settings_label = ctk.CTkLabel(
            master,
            text=label,
        )
        self.settings_label.grid(
            row=row,
            column=0,
            padx=20,
            pady=(10, 0),
        )
        self.settings_value = ctk.StringVar(
            self, self._root().config["files"].get(self.value_key, "")
        )
        self.settings_entry = ctk.CTkEntry(
            master,
            textvariable=self.settings_value,
        )
        self.settings_entry.grid(
            row=row,
            column=1,
            columnspan=2,
            padx=20,
            pady=(10, 0),
            sticky="we",
        )
        self.settings_entry.bind(
            "<Return>",
            self.change_settings_callback,
        )
        self.open_browse_dir_button = ctk.CTkButton(
            master,
            text="Browse",
            command=self.browse_dir if self.setting_is_dir else self.browse_file,
        )
        self.open_browse_dir_button.grid(
            row=row + 1,
            column=1,
            padx=20,
            pady=(5, 10),
            sticky="we",
            columnspan=2,
        )

    def change_settings_callback(self, event=None):
        if self.settings_value.get() == "":
            new_value = ""
        else:
            new_value = fpe(self.settings_value.get())

        self._root().config["files"][self.value_key] = str(new_value)
        self._root().write_config()

        self.master.focus()

    def browse_dir(self):
        dir_output = ctk.filedialog.askdirectory(
            parent=self,
            initialdir=fpe(self.settings_value.get()),
        )
        self.settings_value.set(dir_output)
        self.change_settings_callback()

    def browse_file(self):
        path_output = str(
            ctk.filedialog.askopenfilename(
                parent=self,
                initialdir=fpe(self.settings_value.get()).parent,
            )
        )
        if Path(path_output) is not None and Path(path_output).is_file():
            self.settings_value.set(path_output)
            self.change_settings_callback()


class SettingsWindow(ctk.CTkToplevel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.geometry("720x568")
        self.title("Settings")

        # Key bindings
        self.protocol("WM_DELETE_WINDOW", self.quit_settings_gracefully)
        self.bind("<Control-q>", self.quit_settings_gracefully)
        self.bind("<Control-w>", self.quit_settings_gracefully)

        self.scrollable_frame = ctk.CTkScrollableFrame(self)
        self.scrollable_frame.grid_columnconfigure(0, weight=0)
        self.scrollable_frame.grid_columnconfigure(1, weight=1)
        # self.scrollable_frame.grid_columnconfigure(1, weight=1)
        self.scrollable_frame.pack(side="top", fill="both", expand=True)

        self.rescan_button = ctk.CTkButton(
            self.scrollable_frame,
            text="Rescan directories / Reload",
            command=self._root().rescan_and_reload,
        )
        self.rescan_button.grid(
            row=0,
            column=0,
            padx=20,
            pady=20,
            sticky="we",
            columnspan=2,
        )

        self.appearance_label = ctk.CTkLabel(
            self.scrollable_frame, text="Appearance mode"
        )
        self.appearance_label.grid(
            row=1,
            column=0,
            padx=20,
            pady=20,
        )
        self.change_appearance_menu = ctk.CTkOptionMenu(
            self.scrollable_frame,
            values=["System", "Light", "Dark"],
            command=self.change_appearance_menu_callback,
        )
        self.change_appearance_menu.grid(row=1, column=1, padx=20, pady=20, sticky="w")
        self.change_appearance_menu.set(
            self._root().config["appearance"]["appearance_mode"]
        )
        appearance_info_label = ctk.CTkLabel(
            self.scrollable_frame, text="(Click reload to change the plots)"
        )
        appearance_info_label.grid(
            row=1,
            column=2,
            padx=20,
            pady=20,
        )

        proper_names = [
            "Output directory",
            "Extractions directory",
            "Catalogue filepath",
            "Prep directory",
            "Cube filepath",
            "Temporary directory",
        ]
        backend_names = [
            "out_dir",
            "extractions_dir",
            "cat_path",
            "prep_dir",
            "cube_path",
            "temp_dir",
        ]
        is_dir = [1, 1, 0, 1, 0, 1]

        additional_settings = []

        for i, (p, b, d) in enumerate(zip(proper_names, backend_names, is_dir)):
            additional_settings.append(
                SettingsSelection(
                    self.scrollable_frame,
                    row=2 * (i + 1),
                    label=p,
                    value=b,
                    setting_is_dir=d,
                )
            )

    def change_appearance_menu_callback(self, choice):
        ctk.set_appearance_mode(choice.lower())
        self._root().config["appearance"]["appearance_mode"] = choice.lower()
        self._root().write_config()
        self._root().change_plot_colours()

    def quit_settings_gracefully(self, event=None):
        # Put some lines here to save current output
        self._root().write_config()
        self.destroy()



from pathlib import Path

import customtkinter as ctk

from .base_window import BaseWindow


class CommentsWindow(BaseWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, title="Comments", **kwargs)
        self.top_label.configure(
            text=f"Insert any additional comments for object "
            f"{self._root().current_gal_id.get()} here:"
        )
        self.action_button.configure(text="Save")

    def create_text_box(self):
        super().create_text_box()
        if "comments" in self._root().current_gal_data.keys():
            self.text_box.insert("1.0", self._root().current_gal_data["comments"])

    def action_button_callback(self, event=None):
        self._root().current_gal_data["comments"] = self.text_box.get("1.0", "end")
        self.destroy()



import customtkinter as ctk


class BaseWindow(ctk.CTkToplevel):
    def __init__(self, master, title=None, top_text="", **kwargs):
        super().__init__(master, **kwargs)
        self.geometry("720x568")
        self.title(title)
        self.top_text = top_text

        # Key bindings
        self.protocol("WM_DELETE_WINDOW", self.quit_no_action)
        self.bind("<Control-w>", self.quit_no_action)
        self.bind("<Control-s>", self.action_button_callback)

        self.main_frame = ctk.CTkFrame(self)
        self.main_frame.grid_columnconfigure((0, 1), weight=1)
        self.main_frame.grid_rowconfigure(1, weight=1)
        self.main_frame.pack(side="top", fill="both", expand=True)

        self.create_top_label()
        self.create_text_box()
        self.action_button = ctk.CTkButton(
            self.main_frame,
            text="Action",
            command=self.action_button_callback,
        )
        self.action_button.grid(
            row=2,
            column=0,
            padx=20,
            pady=(5, 10),
        )
        self.close_button = ctk.CTkButton(
            self.main_frame,
            text="Cancel",
            command=self.quit_no_action,
        )
        self.close_button.grid(
            row=2,
            column=1,
            padx=20,
            pady=(5, 10),
        )

    def create_top_label(self):
        self.top_label = ctk.CTkLabel(
            self.main_frame,
            text=self.top_text,
        )
        self.top_label.grid(
            row=0,
            column=0,
            columnspan=2,
            padx=20,
            pady=(10, 0),
        )

    def create_text_box(self):
        self.text_box = ctk.CTkTextbox(
            self.main_frame,
        )

        self.text_box.grid(
            row=1, column=0, padx=20, pady=(10, 0), columnspan=2, sticky="news"
        )
        self.text_box.bind("<Control-Key-a>", self.select_all)
        self.text_box.bind("<Control-Key-A>", self.select_all)

    def select_all(self, event=None):
        self.text_box.tag_add("sel", "1.0", "end")
        self.text_box.mark_set("insert", "1.0")
        self.text_box.see("insert")
        return "break"

    def quit_no_action(self, event=None):
        self.destroy()

    def action_button_callback(self, event=None):
        return



import sys
import tkinter as tk
import warnings
from typing import Callable, Literal, Optional, Tuple, Union

import customtkinter as ctk
from customtkinter.windows.widgets.utility import (
    check_kwargs_empty,
    pop_from_dict_by_set,
)
from PIL import Image, ImageOps


class IconCheckBox(ctk.CTkBaseClass):
    """
    Checkbox with rounded corners. Default is fg_color=None (transparent fg_color).
    """

    # attributes that are passed to and managed by the tkinter entry only:
    _valid_tk_label_attributes = {
        "cursor",
        "justify",
        "padx",
        "pady",
        "textvariable",
        "state",
        "takefocus",
        "underline",
    }

    def __init__(
        self,
        master: any,
        checked_icon: Union[str, ctk.CTkImage] = None,
        unchecked_icon: Union[str, ctk.CTkImage] = None,
        dark_checked_icon: Optional[ctk.CTkImage] = None,  # Add icons for dark theme
        dark_unchecked_icon: Optional[ctk.CTkImage] = None,
        auto_dark_icon: bool = False,  # If dark theme image not given, this will automatically invert the
        # image supplied to create a dark theme image
        width: int = 20,
        height: int = 20,
        corner_radius: Optional[int] = None,
        border_width: Optional[int] = None,
        border_spacing: int = 2,
        on_value: Union[int, str, bool] = 1,
        off_value: Union[int, str, bool] = 0,
        command: Optional[Callable] = None,
        state: Literal["normal", "disabled"] = "normal",
        hover: bool = True,
        bg_color: Union[str, Tuple[str, str]] = "transparent",
        fg_color: Optional[Union[str, Tuple[str, str]]] = None,
        text_color: Optional[Union[str, Tuple[str, str]]] = None,
        hover_color: Optional[Union[str, Tuple[str, str]]] = None,
        border_color: Optional[Union[str, Tuple[str, str]]] = None,
        text_color_disabled: Optional[Union[str, Tuple[str, str]]] = None,
        text: str = "",
        font: Optional[Union[tuple, ctk.CTkFont]] = None,
        variable: Optional[tk.Variable] = None,
        compound: str = "center",
        anchor: str = "center",  # label anchor: center, n, e, s, w
        wraplength: int = 0,
        **kwargs,
    ):
        # transfer basic functionality (_bg_color, size, __appearance_mode, scaling) to ctk.CTkBaseClass
        super().__init__(master=master, bg_color=bg_color, width=width, height=height)

        # color
        self._fg_color = (
            ctk.ThemeManager.theme["CTkLabel"]["fg_color"]
            if fg_color is None
            else self._check_color_type(fg_color, transparency=True)
        )
        self._text_color = (
            ctk.ThemeManager.theme["CTkLabel"]["text_color"]
            if text_color is None
            else self._check_color_type(text_color)
        )
        self._hover_color: Union[str, Tuple[str, str]] = (
            ctk.ThemeManager.theme["CTkButton"]["hover_color"]
            if hover_color is None
            else self._check_color_type(hover_color)
        )
        self._border_color: Union[str, Tuple[str, str]] = (
            ctk.ThemeManager.theme["CTkButton"]["border_color"]
            if border_color is None
            else self._check_color_type(border_color)
        )
        self._text_color_disabled: Union[str, Tuple[str, str]] = (
            ctk.ThemeManager.theme["CTkButton"]["text_color_disabled"]
            if text_color_disabled is None
            else self._check_color_type(text_color_disabled)
        )
        self._original_color: Union[str, Tuple[str, str]] = ctk.ThemeManager.theme[
            "CTkFrame"
        ]["fg_color"]

        # state
        self._state = state
        self._check_state = False

        # hover
        self._hover = hover

        # shape
        self._size = (width, height)
        self._corner_radius: int = (
            ctk.ThemeManager.theme["CTkButton"]["corner_radius"]
            if corner_radius is None
            else corner_radius
        )
        self._corner_radius = min(self._corner_radius, round(self._current_height / 2))
        self._border_width: int = (
            ctk.ThemeManager.theme["CTkButton"]["border_width"]
            if border_width is None
            else border_width
        )
        self._border_spacing: int = border_spacing

        # text
        self._anchor = anchor
        self._text = text
        self._wraplength = wraplength
        self._compound = compound

        # image
        self.dark_checked_icon = self._check_image_type(
            dark_checked_icon, make_compatible=not auto_dark_icon
        )
        self.dark_unchecked_icon = self._check_image_type(
            dark_unchecked_icon, make_compatible=not auto_dark_icon
        )
        self.checked_icon = self._check_image_type(
            checked_icon, make_compatible=not auto_dark_icon
        )
        self.unchecked_icon = self._check_image_type(
            unchecked_icon, make_compatible=not auto_dark_icon
        )

        if self.dark_checked_icon is None:
            if auto_dark_icon:
                self.dark_checked_icon = self.invert_image(self.checked_icon)
        if self.dark_unchecked_icon is None:
            self.dark_unchecked_icon = self.invert_image(self.unchecked_icon)

        try:
            self.checked_icon = ctk.CTkImage(self.checked_icon)
            self.unchecked_icon = ctk.CTkImage(self.unchecked_icon)
        except:
            warnings.warn(
                "Could not convert images to type ctk.CTkImage. Try using a directory instead."
            )

        if isinstance(self.checked_icon, ctk.CTkImage):
            self.checked_icon._dark_image = self.dark_checked_icon
            self.checked_icon.add_configure_callback(self._update_image)
        if isinstance(self.unchecked_icon, ctk.CTkImage):
            self.unchecked_icon._dark_image = self.dark_unchecked_icon
            self.unchecked_icon.add_configure_callback(self._update_image)

        # checked state
        self._checked_state = False
        self._on_value = on_value
        self._off_value = off_value

        # font
        self._font = ctk.CTkFont() if font is None else self._check_font_type(font)
        if isinstance(self._font, ctk.CTkFont):
            self._font.add_size_configure_callback(self._update_font)

        # command
        self._command = command

        # configure grid system (1x1)
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)

        self._canvas = ctk.CTkCanvas(
            master=self,
            highlightthickness=0,
            width=self._apply_widget_scaling(self._desired_width),
            height=self._apply_widget_scaling(self._desired_height),
        )
        self._draw_engine = ctk.DrawEngine(self._canvas)

        self._variable: tk.Variable = variable

        self._label = tk.Label(
            master=self,
            highlightthickness=0,
            padx=0,
            pady=0,
            borderwidth=0,
            anchor=self._anchor,
            compound=self._compound,
            wraplength=self._apply_widget_scaling(self._wraplength),
            text=self._text,
            font=self._apply_font_scaling(self._font),
        )
        self._label.configure(
            **pop_from_dict_by_set(kwargs, self._valid_tk_label_attributes)
        )

        check_kwargs_empty(kwargs, raise_error=True)

        self._create_grid()
        self._create_bindings()
        self._set_cursor()
        self._update_image()
        self._draw()

    def _set_cursor(self):
        if self._cursor_manipulation_enabled:
            if self._state == tk.DISABLED:
                if sys.platform == "darwin":
                    self._canvas.configure(cursor="arrow")
                elif sys.platform.startswith("win"):
                    self._canvas.configure(cursor="arrow")

            elif self._state == tk.NORMAL:
                if sys.platform == "darwin":
                    self._canvas.configure(cursor="pointinghand")
                elif sys.platform.startswith("win"):
                    self._canvas.configure(cursor="hand2")

    def _check_image_type(self, image: any, make_compatible=True):
        """check image type when passed to widget
        :param make_compatible:
        """
        if image is None:
            return image
        elif isinstance(image, str):
            if make_compatible:
                return ctk.CTkImage(Image.open(image), size=self._size)
            else:
                return Image.open(image)
        elif type(image) == Image:
            if make_compatible:
                return ctk.CTkImage(image, size=self._size)
            else:
                return image

        elif isinstance(image, ctk.CTkImage):
            return image
        else:
            warnings.warn(
                f"{type(self).__name__} Warning: Given image is not ctk.CTkImage but {type(image)}. Image can not be "
                f"scaled on HighDPI displays, use ctk.CTkImage or directory instead.\n"
            )
            return image

    def _create_bindings(self, sequence: Optional[str] = None):
        """set necessary bindings for functionality of widget, will overwrite other bindings"""
        if sequence is None or sequence == "<Enter>":
            self.bind("<Enter>", self._on_enter)
        if sequence is None or sequence == "<Leave>":
            self.bind("<Leave>", self._on_leave)
        if sequence is None or sequence == "<Button-1>":
            self.bind("<Button-1>", self.toggle)

    def _on_enter(self, event=0):
        if self._hover is True and self._state == tk.NORMAL:
            if self._cursor_manipulation_enabled:
                self._label.config(cursor="hand2")
            self._label.config(bg=self._apply_appearance_mode(self._hover_color))
            if self._check_state is True:
                self._canvas.itemconfig(
                    "inner_parts",
                    fill=self._apply_appearance_mode(self._hover_color),
                    outline=self._apply_appearance_mode(self._hover_color),
                )
                self._canvas.itemconfig(
                    "border_parts",
                    fill=self._apply_appearance_mode(self._hover_color),
                    outline=self._apply_appearance_mode(self._hover_color),
                )
            else:
                self._canvas.itemconfig(
                    "inner_parts",
                    fill=self._apply_appearance_mode(self._hover_color),
                    outline=self._apply_appearance_mode(self._hover_color),
                )

    def _on_leave(self, event=0):
        if self._fg_color == "transparent":
            fg_color = self._bg_color
        else:
            fg_color = self._fg_color
        if self._check_state is True:
            if self._fg_color == "transparent":
                inner_parts_color = self._bg_color
            else:
                inner_parts_color = self._fg_color

            self._canvas.itemconfig(
                "inner_parts",
                fill=self._apply_appearance_mode(inner_parts_color),
                outline=self._apply_appearance_mode(inner_parts_color),
            )
            self._canvas.itemconfig(
                "border_parts",
                fill=self._apply_appearance_mode(self._fg_color),
                outline=self._apply_appearance_mode(self._fg_color),
            )
            self._label.config(
                cursor="", bg=self._apply_appearance_mode(inner_parts_color)
            )
        else:
            self._canvas.itemconfig(
                "inner_parts",
                fill=self._apply_appearance_mode(self._bg_color),
                outline=self._apply_appearance_mode(self._bg_color),
            )
            self._canvas.itemconfig(
                "border_parts",
                fill=self._apply_appearance_mode(self._border_color),
                outline=self._apply_appearance_mode(self._border_color),
            )
            self._label.config(
                cursor="", bg=self._apply_appearance_mode(self._bg_color)
            )

    def toggle(self, event=0):
        if self._state == tk.NORMAL:
            if self._checked_state is True:
                self._checked_state = False
                self._update_image()
                self._draw()
            else:
                self._checked_state = True
                self._update_image()
                self._draw()

            if self._variable is not None:
                self._variable_callback_blocked = True
                self._variable.set(
                    self._on_value if self._check_state is True else self._off_value
                )
                self._variable_callback_blocked = False

            if self._command is not None:
                self._command()

            self._on_enter()

    def select(self, from_variable_callback=False):
        self._check_state = True
        self._draw()

        if self._variable is not None and not from_variable_callback:
            self._variable_callback_blocked = True
            self._variable.set(self._on_value)
            self._variable_callback_blocked = False

    def deselect(self, from_variable_callback=False):
        self._check_state = False
        self._draw()

        if self._variable is not None and not from_variable_callback:
            self._variable_callback_blocked = True
            self._variable.set(self._off_value)
            self._variable_callback_blocked = False

    def _set_scaling(self, *args, **kwargs):
        super()._set_scaling(*args, **kwargs)

        self._canvas.configure(
            width=self._apply_widget_scaling(self._desired_width),
            height=self._apply_widget_scaling(self._desired_height),
        )
        self._label.configure(font=self._apply_font_scaling(self._font))
        self._label.configure(wraplength=self._apply_widget_scaling(self._wraplength))

        self._create_grid()
        self._update_image()
        self._draw(no_color_updates=True)

    def _set_appearance_mode(self, mode_string):
        super()._set_appearance_mode(mode_string)
        self._update_image()

    def _set_dimensions(self, width=None, height=None):
        super()._set_dimensions(width, height)
        self._canvas.configure(
            width=self._apply_widget_scaling(self._desired_width),
            height=self._apply_widget_scaling(self._desired_height),
        )
        self._create_grid()
        self._draw()

    def _update_font(self):
        """pass font to tkinter widgets with applied font scaling and update grid with workaround"""
        self._label.configure(font=self._apply_font_scaling(self._font))

        # Workaround to force grid to be resized when text changes size.
        # Otherwise, grid will lag and only resizes if other mouse action occurs.
        self._canvas.grid_forget()
        self._canvas.grid(row=0, column=0, sticky="nsew")

    def _update_image(self):
        new_image = self.checked_icon if self._checked_state else self.unchecked_icon

        if isinstance(new_image, ctk.CTkImage):
            self._label.configure(
                image=new_image.create_scaled_photo_image(
                    self._get_widget_scaling(), self._get_appearance_mode()
                )
            )
        elif new_image is not None:
            self._label.configure(image=new_image)

    def destroy(self):
        if isinstance(self._font, ctk.CTkFont):
            self._font.remove_size_configure_callback(self._update_font)
        super().destroy()

    def get(self):
        return self._on_value if self._checked_state else self._off_value

    def _create_grid(self):
        """configure grid system (1x1)"""

        text_label_grid_sticky = self._anchor if self._anchor != "center" else ""
        self._label.grid(
            row=0,
            column=0,
            sticky=text_label_grid_sticky,
            padx=self._apply_widget_scaling(
                min(self._corner_radius, round(self._current_height / 2))
            ),
        )
        self._canvas.grid(row=0, column=0, sticky="nsew")

    def _draw(self, no_color_updates=False):
        super()._draw(no_color_updates)
        requires_recoloring = self._draw_engine.draw_rounded_rect_with_border(
            self._apply_widget_scaling(self._current_width),
            self._apply_widget_scaling(self._current_height),
            self._apply_widget_scaling(self._corner_radius),
            0,
        )

        if no_color_updates is False or requires_recoloring:
            if self._apply_appearance_mode(self._fg_color) == "transparent":
                self._canvas.itemconfig(
                    "inner_parts",
                    fill=self._apply_appearance_mode(self._bg_color),
                    outline=self._apply_appearance_mode(self._bg_color),
                )

                self._label.configure(
                    fg=self._apply_appearance_mode(self._text_color),
                    bg=self._apply_appearance_mode(self._bg_color),
                )
            else:
                if self._checked_state:
                    self._canvas.itemconfig(
                        "inner_parts",
                        fill=self._apply_appearance_mode(self._fg_color),
                        outline=self._apply_appearance_mode(self._fg_color),
                    )

                    self._canvas.itemconfig(
                        "border_parts",
                        fill=self._apply_appearance_mode(self._border_color),
                        outline=self._apply_appearance_mode(self._border_color),
                    )

                    self._label.configure(
                        fg=self._apply_appearance_mode(self._text_color),
                        bg=self._apply_appearance_mode(self._fg_color),
                    )

                else:
                    self._canvas.itemconfig(
                        "inner_parts",
                        fill=self._apply_appearance_mode(self._bg_color),
                        outline=self._apply_appearance_mode(self._bg_color),
                    )
                    self._canvas.itemconfig(
                        "border_parts",
                        fill=self._apply_appearance_mode(self._border_color),
                        outline=self._apply_appearance_mode(self._border_color),
                    )

                    self._label.configure(
                        fg=self._apply_appearance_mode(self._text_color),
                        bg=self._apply_appearance_mode(self._bg_color),
                    )

            self._canvas.configure(bg=self._apply_appearance_mode(self._bg_color))

    def configure(self, require_redraw=False, **kwargs):
        if "state" in kwargs:
            self._state = kwargs.pop("state")
            self._set_cursor()
            require_redraw = True

        if "corner_radius" in kwargs:
            self._corner_radius = kwargs.pop("corner_radius")
            self._create_grid()
            require_redraw = True

        if "fg_color" in kwargs:
            self._fg_color = self._check_color_type(
                kwargs.pop("fg_color"), transparency=True
            )
            require_redraw = True

        if "text_color" in kwargs:
            self._text_color = self._check_color_type(kwargs.pop("text_color"))
            require_redraw = True

        if "text" in kwargs:
            self._text = kwargs.pop("text")
            self._label.configure(text=self._text)

        if "font" in kwargs:
            if isinstance(self._font, ctk.CTkFont):
                self._font.remove_size_configure_callback(self._update_font)
            self._font = self._check_font_type(kwargs.pop("font"))
            if isinstance(self._font, ctk.CTkFont):
                self._font.add_size_configure_callback(self._update_font)
            self._update_font()

        if "compound" in kwargs:
            self._compound = kwargs.pop("compound")
            self._label.configure(compound=self._compound)

        if "anchor" in kwargs:
            self._anchor = kwargs.pop("anchor")
            self._label.configure(anchor=self._anchor)
            self._create_grid()

        if "wraplength" in kwargs:
            self._wraplength = kwargs.pop("wraplength")
            self._label.configure(
                wraplength=self._apply_widget_scaling(self._wraplength)
            )

        if "checked_icon" in kwargs:
            self.checked_icon = kwargs.pop("checked_icon")
            require_redraw = True

        if "unchecked_icon" in kwargs:
            self.checked_icon = kwargs.pop("unchecked_icon")
            require_redraw = True

        self._label.configure(
            **pop_from_dict_by_set(kwargs, self._valid_tk_label_attributes)
        )  # configure tkinter.Label

        super().configure(
            require_redraw=require_redraw, **kwargs
        )  # configure ctk.CTkBaseClass

    def cget(self, attribute_name: str) -> any:
        if attribute_name == "corner_radius":
            return self._corner_radius

        elif attribute_name == "fg_color":
            return self._fg_color
        elif attribute_name == "text_color":
            return self._text_color

        elif attribute_name == "text":
            return self._text
        elif attribute_name == "font":
            return self._font
        elif attribute_name == "checked_icon":
            return self.checked_icon
        elif attribute_name == "unchecked_icon":
            return self.unchecked_icon
        elif attribute_name == "compound":
            return self._compound
        elif attribute_name == "anchor":
            return self._anchor
        elif attribute_name == "wraplength":
            return self._wraplength

        elif attribute_name in self._valid_tk_label_attributes:
            return self._label.cget(attribute_name)  # cget of tkinter.Label
        else:
            return super().cget(attribute_name)  # cget of ctk.CTkBaseClass

    def bind(self, sequence: str = None, command: Callable = None, add: str = True):
        """called on the tkinter.Label and tkinter.Canvas"""
        if not (add == "+" or add is True):
            raise ValueError(
                "'add' argument can only be '+' or True to preserve internal callbacks"
            )
        self._canvas.bind(sequence, command, add=True)
        self._label.bind(sequence, command, add=True)

    def unbind(self, sequence: str = None, funcid: Optional[str] = None):
        """called on the tkinter.Label and tkinter.Canvas"""
        if funcid is not None:
            raise ValueError(
                "'funcid' argument can only be None, because there is a bug in"
                + " tkinter and its not clear whether the internal callbacks will be unbinded or not"
            )
        self._canvas.unbind(sequence, None)
        self._label.unbind(sequence, None)

    def focus(self):
        return self._label.focus()

    def focus_set(self):
        return self._label.focus_set()

    def focus_force(self):
        return self._label.focus_force()

    def invert_image(self, image) -> Image:
        """Invert an image and return the inverted image"""
        try:
            if image.mode == "RGBA":
                r, g, b, a = image.split()
                rgb_image = Image.merge("RGB", (r, g, b))
                inverted_image = ImageOps.invert(rgb_image)
                r2, g2, b2 = inverted_image.split()
                final_transparent_image = Image.merge("RGBA", (r2, g2, b2, a))
                return final_transparent_image
            else:
                inverted_image = ImageOps.invert(image)
                return inverted_image
        except AttributeError as e:
            warnings.warn(
                f"{e}\nAn incompatible file type was probably used. Use PIL to open a file instead."
            )
            return image
        except Exception as e:
            warnings.warn(f"{e}\nAn unknown error is preventing image inversion.")
            return image
